import 'dart:async';

import 'package:chopper/chopper.dart';
import 'package:libaqueous/src/api/api.dart';
import 'package:libaqueous/src/api/events/model/event.dart';
import 'package:libaqueous/src/api/room/model/room_member.dart';
import 'package:libaqueous/src/api/sync/model/invited_room_sync.dart';
import 'package:libaqueous/src/api/sync/model/room_sync.dart';
import 'package:libaqueous/src/api/sync/model/sync_response.dart';
import 'package:libaqueous/src/client/event_bus.dart';
import 'package:libaqueous/src/client/events.dart';
import 'package:libaqueous/src/client/room/room.dart';
import 'package:libaqueous/src/client/room/room_account_data.dart';
import 'package:libaqueous/src/client/room/room_state.dart';
import 'package:libaqueous/src/client/room/room_timeline.dart';
import 'package:libaqueous/src/client/store/store.dart';
import 'package:libaqueous/src/client/user/user.dart';
import 'package:libaqueous/src/client/user/user_state.dart';

class Client {
  Client(String homeserver, userID, token, this.store)
      : api = MatrixAPI(homeserver, userID, token),
        id = DateTime
            .now()
            .millisecondsSinceEpoch;

  String get homeServer => api.baseUrl;

  String get userID => api.user;

  String get token => api.token;

  final int id;

  int _transactionCount = 0;

  String get transactionID =>
      id.toString() + "q" + (++_transactionCount).toString();

  MatrixAPI api;

  bool _running = false;

  Store store;

  Map<String, Room> rooms = {};
  Map<String, User> users = {};

  Future<void> loadFromStore() async {
    final roomIDs = await store.roomIDs();
    if (roomIDs?.isEmpty ?? true) return;
    for (final id in roomIDs) {
      final room = await _roomOrCreate(id);
    }

    final userIDs = await store.userIDs();
    if (userIDs?.isEmpty ?? true) return;
    for (final id in userIDs) {
      final user = await _userOrCreate(id);
    }
  }

  void startSync() {
    if (_running) return;
    _running = true;

    _sync();
  }

  void stopSync() {
    _running = false;
  }

  Future<void> _sync() async {
    var isInitialSync = (await store.syncToken())?.isEmpty ?? true;

    final stream = _getEventStream();

    await for (SyncResponse data in stream) {
      await _processSyncResponse(data);

      if (isInitialSync) {
        isInitialSync = false;
        EventHandler.eventBus.fire(InitialSyncCompleted());
      }
    }
  }

  Stream<SyncResponse> _getEventStream() async* {
    while (_running) {
      final syncToken = await store.syncToken() ?? "";
      print("Previous token: " + syncToken);

      if (syncToken.isEmpty) {
        print("Fetching full state...");
      }

      final Response<SyncResponse> resp = await api.syncAPI.sync(
          "{\"room\":{\"state\":{\"lazy_load_members\":true}}}",
          syncToken,
          syncToken.isEmpty,
          30000);
      final syncResp = resp.body;

      yield syncResp;

      await store.setSyncToken(syncResp.nextBatch);
    }
  }

  Future<void> _processSyncResponse(SyncResponse data) async {
    if (data == null) return;

    // TODO: Process and store sync response.
    if (data.toDevice?.events?.isNotEmpty ?? false) {
      for (var e in data.toDevice.events) {}
    }

    if (data.accountData?.events?.isNotEmpty ?? false) {
      for (var e in data.accountData.events) {}
    }

    if (data.presence?.events?.isNotEmpty ?? false) {
      for (Event presenceEvent in data.presence.events) {
        final user = _userOrCreate(presenceEvent.sender);
      }
    }

    if (data.rooms?.join?.isNotEmpty ?? false) {
      await _processRooms(data.rooms.join, "join");
    }

    if (data.rooms?.leave?.isNotEmpty ?? false) {
      await _processRooms(data.rooms.leave, "leave");
    }

    if (data.rooms?.invite?.isNotEmpty ?? false) {
      await _processInvitedRooms(data.rooms.invite);
    }
  }

  Future<void> _processRooms(
      Map<String, RoomSync> roomSyncMap, String membership) async {
    for (var id in roomSyncMap.keys) {
      var roomSync = roomSyncMap[id];

      int highlightCount = roomSync.unreadNotifications?.highlightCount ?? 0;
      int notificationCount =
          roomSync.unreadNotifications?.notificationCount ?? 0;
      String prevToken = roomSync.timeline?.prevToken ?? "";
      bool limited = roomSync.timeline?.limited ?? false;

      var room = await _roomOrCreate(id)
        ..state.userMembership = membership;

      if (notificationCount != room.state.notificationCount) {
        EventHandler.eventBus
            .fire(RoomNotificationCountChanged(notificationCount));
      }

      await _processRoomStateEvents(id, roomSync.state?.events);
      await _processRoomTimelineEvents(id, roomSync.timeline?.events);
    }
  }

  Future<void> _processInvitedRooms(
      Map<String, InvitedRoomSync> roomSyncMap) async {
    for (var id in roomSyncMap.keys) {
      var roomSync = roomSyncMap[id];

      var room = await _roomOrCreate(id);

      room.state.userMembership = RoomMember.MEMBERSHIP_INVITE;
    }
  }

  Future<void> _processRoomStateEvents(String id, List<Event> events) async {
    var room = await _roomOrCreate(id);

    await room.processStateEvents(events);

    await store.storeRoom(room);
  }

  Future<void> _processRoomTimelineEvents(String id, List<Event> events) async {
    var room = await _roomOrCreate(id);

    await room.processTimelineEvents(events, false);

    await store.storeRoom(room);
  }

  Future<Room> _roomOrCreate(String id) async {
    if (!rooms.containsKey(id)) {
      final room = Room(
          this,
          await store.roomState(id) ?? RoomState()
            ..roomID = id,
          await store.roomAccountData(id) ?? RoomAccountData(),
          await store.roomTimeline(id) ?? RoomTimeline());
      rooms[id] = room;
      EventHandler.eventBus.fire(NewRoom(room));
      store.storeRoom(room);
    }
    return rooms[id];
  }

  Future<User> _userOrCreate(String id) async {
    if (!users.containsKey(id)) {
      final user = User(await store.userState(id) ?? UserState()
        ..userID = id);
      users[id] = user;
      EventHandler.eventBus.fire(NewUser(user));
      store.storeUser(user);
    }
    return users[id];
  }
}
