import 'package:libaqueous/src/api/events/model/event.dart';
import 'package:libaqueous/src/api/events/model/event_type.dart';
import 'package:libaqueous/src/client/client.dart';
import 'package:libaqueous/src/client/event_bus.dart';
import 'package:libaqueous/src/client/events.dart';
import 'package:libaqueous/src/client/room/room_account_data.dart';
import 'package:libaqueous/src/client/room/room_state.dart';
import 'package:libaqueous/src/client/room/room_timeline.dart';
import 'package:libaqueous/src/client/user/user.dart';
import 'package:libaqueous/src/client/user/user_state.dart';

class Room {
  Room(this.client, this.state, this.accountData, this.timeline);

  String get id => state.roomID;

  String get displayName => state.name ?? state.canonicalAlias ?? state.roomID;

  final RoomState state;
  final RoomAccountData accountData;
  final RoomTimeline timeline;

  final Client client;

  Future<void> processStateEvents(List<Event> events) async {
    for (var event in events) {
      await processStateEvent(event);
    }
    EventHandler.eventBus.fire(RoomStateChanged(this, state));
  }

  Future<void> processStateEvent(Event event) async {
    switch (event.type) {
      case EventType.STATE_ROOM_NAME:
        state.name = event.content["name"];
        break;
      case EventType.STATE_ROOM_TOPIC:
        state.topic = event.content["topic"];
        break;
      case EventType.STATE_ROOM_AVATAR:
        state.avatarUrl = Uri.parse(event.content["url"] ?? "");
        break;
      case EventType.STATE_CANONICAL_ALIAS:
        state.canonicalAlias = event.content["alias"];
        break;
      case EventType.STATE_ROOM_MEMBER:
        if (!client.users.containsKey(id)) {
          final user = User(await client.store.userState(id) ?? UserState()
            ..userID = id
            ..name = event.content["displayname"]
            ..avatarUrl = Uri.parse(event.content["avatar_url"] ?? ""));
          client.users[id] = user;
          EventHandler.eventBus.fire(NewUser(user));
          client.store.storeUser(user);
        } else {
          final user = client.users[id];

          user.state
            ..avatarUrl = event.content.containsKey("avatar_url") &&
                (event.content["avatar_url"]
                    ?.toString()
                    ?.isNotEmpty ?? false)
                ? Uri.parse(event.content["avatar_url"])
                : null
            ..name = event.content["displayname"];

          client.store.storeUser(user);
        }
    }
  }

  Future<void> processTimelineEvents(List<Event> events, bool isHistory) async {
    bool processedStateEvent = false;

    for (var event in events) {
      if (EventType.isStateEvent(event.type)) {
        processedStateEvent = true;
        await processStateEvent(event);
      }
      await processTimelineEvent(event, isHistory);
    }

    if (processedStateEvent) {
      EventHandler.eventBus.fire(RoomStateChanged(this, state));
    }
  }

  Future<void> processTimelineEvent(Event event, bool isHistory) async {
    if (isHistory) {
      timeline.events.insert(0, event);
      EventHandler.eventBus.fire(RoomHistoricalEvent(this, event));
    } else {
      timeline.events.add(event);
      EventHandler.eventBus.fire(RoomNewEvent(this, event));
    }
  }

  Future<void> postMessage(Map<String, dynamic> json) async {
    await client.api.roomAPI
        .send(this.id, "m.room.message", client.transactionID, json);
  }

  Future<void> postPlainText(String text) async {
    await postMessage({"msgtype": "m.text", "body": text});
  }

  Future<void> postEmote(String text) async {
    await postMessage({"msgtype": "m.emote", "body": text});
  }
}
